string cvs_version = "$Id: install-roxen.pike,v 1.11 2003/06/17 17:18:14 grubba Exp $";

Tools.Install.ProgressBar progress_bar;
Tools.Install.Readline rl;
Stdio.File infd, outfd;
Stdio.Terminfo.Termcap term;

// Taken from Roxen/2.1/server/etc/modules/Roxen.pmod.
string get_world(array(string) urls) 
{
  if(!sizeof(urls)) 
    return 0;

  string url=urls[0];
  foreach( ({"http:","fhttp:","https:","ftp:"}), string p)
    foreach(urls, string u)
      if(u[0..sizeof(p)-1]==p) {
	url=u;
	break;
      }

  string protocol, server, path="";
  int port;
  if(sscanf(url, "%s://%s:%d/%s", protocol, server, port, path)!=4 &&
     sscanf(url, "%s://%s:%d", protocol, server, port)!=3 &&
     sscanf(url, "%s://%s/%s", protocol, server, path)!=3 &&
     sscanf(url, "%s://%s", protocol, server)!=2 )
    return 0;

  if(protocol=="fhttp") protocol="http";

  array hosts=({ gethostname() }), dns;
  catch(dns=Protocols.DNS.client()->gethostbyname(hosts[0]));
  if(dns && sizeof(dns))
    hosts+=dns[2]+dns[1];

  foreach(hosts, string host)
    if(glob(server, host)) 
    {
      server=host;
      break;
    }

  // Strip extra path info from URL.
  path = (path/"#")[0];

  if(port) 
    return sprintf("%s://%s:%d/%s", protocol, server, port, path);
  return sprintf("%s://%s/%s", protocol, server, path);
}

string read_admin_url(string dir)
{
  foreach(get_dir(dir) || ({}), string filename)
  {
    filename = combine_path(dir, filename);
    if(Stdio.is_file(filename))
    {
      string s = Stdio.read_file(filename);

      if(s)
      {
	string url;
	
	if(sscanf(s, "%*sconfig_filesystem%*s<var name='URLs'> <a> <str>%s</str></a> </var>", url) == 3)
	  return url;
      }
    }
  }

  return 0;
}

void fatal(string fmt, mixed ... argv)
{
  write("\n   FATAL: "+sprintf(fmt, @argv)+"\n");
  exit(1);
}

void bootstrap(string path, mapping opt, array(string) argv)
{
  foreach(sort(glob("*.pike", get_dir(path) || ({}))), string script)
  {
    object o = ((program)combine_path(path, script))(opt);
    
    if(o->main)
      o->main(1 + sizeof(argv[1..]), ({ script }) + argv[1..]);
  }
}

void say_welcome()
{
  write("\n   Welcome to the interactive Roxen 4.0 installation script.\n"
	"\n"
	"   This script takes a number of options; use --help to list them.\n"
	"\n");
  if(rl->get_input_controller()->dumb)
      write("   The script will guide you through the installation process "
	    "by asking\n   a few questions.\n\n");
  else
      write("   The script will guide you through the installation process "
	    "by asking\n   a few questions. Whenever you input a path or a "
	    "filename, you may use\n   the <tab> key to perform filename "
	    "completion.\n\n");
}

int term_height()
{
  catch
  {
    int r = outfd->tcgetattr()->rows;
    if(r)
      return r;
  };
  return term->tgetnum("li") || 24;
}

void press_space()
{
  Stdio.Readline.InputController ic = 
    Stdio.Readline.InputController(outfd, term);
  Stdio.Readline.OutputController oc = 
    Stdio.Readline.OutputController(infd, term);
  ic->nullbindings(); 
  ic->bind(" ", ic->disable); 
  ic->bind("\n", ic->disable); 
  oc->turn_on("reverse");
  oc->write(">> Press space or enter <<"); 
  oc->turn_off("reverse");
  ic->run_blocking();
  write("\r"+" "*78+"\r");
}

string ask_installation_prefix(string path, string prefix)
{
  if(sizeof(prefix))
  {
    path = rl->absolute_path(prefix);
    write("   Installation prefix is %s.\n", path);
  }
  else
    path = rl->absolute_path
      (rl->edit_directory
       (path, "Installation prefix (\"roxen\" will be added): ") || path);
  return path+"/roxen";
}

void shift_old_installations(string path)
{
  // Make sure no old junk is left.
  Stdio.recursive_rm(path+"/server.tmp");
  // These moves are obsolete with versioned server directories.
  Stdio.recursive_rm(path+"/server.older");
  mv(path+"/server.old", path+"/server.older");
  mv(path+"/server", path+"/server.old");
}

array(array) extra_args(array(string) argv)
  // Separate arguments after "--".
{
  array args = ((argv+({ "--" }))/({ "--" }))[0..1];
  return ({ args[0], ({ "install" }) + args[1] });
}

string install_packages(string roxen_path)
{
  write("\n   Installing packages, please wait...\n\n");

  string original_dir = getcwd();

  string package_path = combine_path(original_dir, "..");
  array(string) files = glob("*.tar.gz", get_dir(package_path));
  progress_bar = Tools.Install.ProgressBar("Installing", 0, sizeof(files));
  foreach(files, string file)
  {
    cd(package_path);
    progress_bar->update(1);
    string package = file-".tar.gz";
    sscanf(file, "%s-%*s.tar.gz", package);
    if(!cd(roxen_path))
      fatal("Cannot cd to %s.", roxen_path);
    string r =
      Process.popen("(gzip -d <"+package_path+"/"+file+" | tar xf -) 2>&1");
    if(r && sizeof(r) && 
       // Problem with MacOS X.
       !glob("*tar: End of archive volume 1 reached*", r))
      fatal("%s\n", r);
  }
  write("\r"+" "*78+"\r"+
	"   Packages installation completed successfully.\n\n");

  string ver = Stdio.read_file(roxen_path+"/server.tmp/VERSION");
  if(!ver)
    fatal("Cannot read %O.", roxen_path+"/server.tmp/VERSION");
  sscanf(ver, "%s\n", ver);
  string server_dir = "server-"+ver;
  if(!mv(roxen_path+"/server.tmp", roxen_path+"/"+server_dir))
    fatal("Cannot mv %O to %O. Maybe this version alreay is installed?", 
	  roxen_path+"/server.tmp", roxen_path+"/"+server_dir);

  // Remove environment file which may have paths to the old server
  // directory.
  mv(roxen_path+"/local/environment", roxen_path+"/local/environment~");
  
  cd(original_dir);
  return server_dir;
}

void install_pike(string roxen_path, string server_dir, 
		  array(string) argv, array rest)
{
  string pike_path = roxen_path+"/"+server_dir+"/pike";
  argv += ({ "prefix="+pike_path, "--no-x", "--traditional" });
  program p = (program)(getcwd()+"/bin/install.pike");
  if(p)
    p()->main(sizeof(argv), copy_value(argv), @rest);
  else
    fatal("Pike installation not found.");
  if(!file_stat(pike_path+"/bin/pike"))
    fatal("Pike binary not found.");

  string real_pike = pike_path+"/bin/pike";
  if(Stdio.is_link(real_pike))
    real_pike = readlink(pike_path+"/bin/pike");
  string roxen_pike = roxen_path+"/"+server_dir+"/bin/roxen";
  mixed err = catch
  {
    rm(roxen_pike);
    hardlink(real_pike, roxen_pike);
  };
  if(err)
    err = catch 
    {
      rm(roxen_pike);
      // FIXME: It is probably better to do a copy here instead.
      symlink(pike_path+"/bin/pike", roxen_pike);
    };
  if(err)
    fatal("Unable to link pike binary!\n");
}

void install_config_interface(string roxen_path)
{
  write("\n");
  Process.create_process(({ "./create_configinterface" }))->wait();
  if(!Stdio.is_dir(roxen_path+"/configurations"))
    fatal("Configuration of administration interface FAILED.");
}

void run_bootstraps(string roxen_path, array argv, array eargv)
{
  // Run scripts which possibly have been added by 3:rd party components.
  bootstrap(roxen_path+"/var/bootstrap",
	    ([
	      "roxen_path":roxen_path
	    ]),
	    argv + eargv);
}

void start_roxen_and_wait(string roxen_path)
{
  write("\n   The initial installation is complete. Starting Roxen,\n"
	"   please wait...");
  Process.create_process(({ "./start", "--silent-start" }))->wait();

  int success = 0;
  for(int tryout = 0; tryout < 60; tryout++)
  {
    string lf =
      Stdio.read_file(combine_path(roxen_path, "logs/debug/default.1")) || "";
    foreach(lf/"\n", string row)
    {
      // Wait for the first start identification row. Format:
      // "          : ** 2000-09-01 20:14   pid: 0   ppid: 0   uid: guru"
      if(glob("* : *-*-* *:* pid:* ppid:* uid:*", row))
      {
	success = 1;
	break;
      }
    }
    if(success)
      break;
    sleep(2);
    write(".");
  }

  if(success)
  {
    string adminurl = read_admin_url(roxen_path + "/configurations");
    write("\n\n"
	  "   Congratulations! Roxen has started. You may now proceed with the\n"
	  "   installation from the administration interface using a web browser"
	  
	  +(adminurl ? ":\n\n"+"   "+get_world(({ adminurl })) : ".")+"\n\n");
  }
  else
  {
    write("\n\n"
	  "   WARNING: It seems the server did not start properly (within\n"
	  "   120 seconds). Please consult the debug log file located at\n   "+
	  combine_path(roxen_path, "logs", "debug", "default.1")+
	  " for details.\n");
  }
}

void start_roxen_self_test(string roxen_path)
{
  write("\n   The initial installation is complete. "
	"Starting Roxen self test.\n\n");
  Process.create_process(({ "./start", "--self-test" }))->wait();
}

void no_start()
{
    write("\n   The initial installation is complete. Roxen is not started\n"
	  "   automatically.\n");
}

int ask_start_now()
{
  write("\n");
  string r;
  do 
    r = lower_case
	(rl->edit("", "Would you like to start the server now [Y/n]? "));
  while(r != "" && r != "y" && r != "n");
  return r == "y" || r == "";
}

int main(int argc, array(string) argv, mixed ... rest)
{
  [argv, array(string) eargv] = extra_args(argv);
  
  int auto_license = Getopt.find_option(eargv, "y");
  int is_self_test = 0;
  string default_prefix = "";
  if(Getopt.find_option(eargv, 0, "self-test"))
  {
    is_self_test = 1;
    default_prefix = ".";
  }
  
  infd = Stdio.stdin;
  outfd = Stdio.stdout;
  term = Stdio.Terminfo.getTerm();
  rl = Tools.Install.Readline(0, term, outfd);
  rl->set_cwd("../..");
  
  if(!is_self_test)
  {
    say_welcome();
  }
  string roxen_path = ask_installation_prefix
    ("/usr/local", Getopt.find_option(eargv, 0, "prefix", 0, default_prefix));
  if (!Stdio.mkdirhier(roxen_path)) {
    fatal("Could not create directory %O.", roxen_path);
  }
  shift_old_installations(roxen_path);
  string server_dir = install_packages(roxen_path);
  install_pike(roxen_path, server_dir, argv, rest);
  
  cd(roxen_path+"/"+server_dir);
  
  if(!is_self_test)
    install_config_interface(roxen_path);
  run_bootstraps(roxen_path, argv, eargv);
  
  if(is_self_test)
    start_roxen_self_test(roxen_path);
  else if(Getopt.find_option(eargv, 0, "without-start") || !ask_start_now())
    no_start();
  else
    start_roxen_and_wait(roxen_path);
}
